/**
  *  \file client/wedit.cc
  *
  *  (c) 2001-2006 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cctype>
#include <algorithm>
#include "client/wedit.h"
#include "cpluslib/assert.h"
#include "util/recode.h"
#include "ui/skin.h"
#include "ui/uiglobal.h"
#include "util/console.h"

WEditor::WEditor(int id)
    : UIWidget(id), max_line_length(string_t().max_size()),
      min_line(0), max_line(content.max_size()), max_storage(content.max_size()),
      cursor_x(0), cursor_y(0)
{ }

WEditor::~WEditor()
{ }

void
WEditor::setMaxLineLength(col_t len)
{
    max_line_length = len;
}

void
WEditor::setMaxStoredLines(line_t lines)
{
    max_storage = lines;
    if (max_storage < content.size())
        content.erase(content.begin() + max_storage, content.end());
}

void
WEditor::setUserLineRestrictions(line_t min, line_t max)
{
    ASSERT(min < max);
    min_line = min;
    max_line = max;
    if (cursor_y < min_line)
        cursor_y = min_line;
    if (cursor_y >= max_line)
        cursor_y = max_line-1;
}

void
WEditor::insertLine(line_t line)
{
    if (line >= content.size())
        return;
    if (content.size() < max_storage)
        content.push_back(string_t());
    for (line_t i = content.size()-1; i > line; --i)
        content[i] = content[i-1];
    content[line] = string_t();
    drawWidget();
}

void
WEditor::deleteLine(line_t line, line_t count)
{
    /* we have situation:
         [0, line) [line, line+count) [line+count, size)
       and want
         [0, line) [line+count, size) */
    if (count == 0) {
        /* nothing */
    } else if (content.size() < line + count) {
        if (content.size() > line)
            content.erase(content.begin() + line, content.end());
    } else {
        vec_t::iterator dest = content.begin() + line,
            src = dest + count;
        while (src != content.end()) {
            *dest = *src;
            ++src, ++dest;
        }
        content.erase(dest, content.end());
    }
    drawWidget();
}

string_t
WEditor::getLine(line_t line) const
{
    if (line < content.size())
        return content[line];
    else
        return string_t();
}

void
WEditor::setLine(line_t line, string_t text)
{
    if (line >= content.size()) {
        if (!text.empty()) {
            content.insert(content.end(), line - content.size(), string_t());
            content.push_back(text);
            drawWidget();
        }
    } else {
        if (text != content[line]) {
            content[line] = text;
            drawWidget();
        }
    }
}

string_t
WEditor::getContentAsString() const
{
    return getContentAsString(0, max_line);
}

string_t
WEditor::getContentAsString(line_t begin, line_t end) const
{
    string_t rv;
    if (end > content.size())
        end = content.size();
    for (line_t i = begin; i != end; ++i) {
        rv += content[i];
        rv += '\n';
    }
    return rv;
}

static bool
canBreakLineAt(char c)
{
    using std::isspace;
    return isspace(static_cast<unsigned char>(c))
        || c == '-' || c == '/';
}

void
WEditor::splitLineNoRedraw(line_t line, col_t col)
{
    if (line < content.size()) {
        insertLine(line + 1);
        if (content[line].size() > col) {
            setLine(line + 1, strRTrim(content[line].substr(col)));
            setLine(line,     strRTrim(content[line].substr(0, col)));
        }
    }

    if (cursor_y == line && cursor_x >= col) {
        ++cursor_y;
        cursor_x -= col;
    }
}

void
WEditor::wrapLineNoRedraw(line_t line, col_t length)
{
    if (length <= 0)
        return;                 // avoid stupid case

    while (line < content.size() && content[line].length() > length) {
        col_t split_at = length;
        for (col_t i = length; i > 1; --i) {
            if (canBreakLineAt(content[line][i-1])) {
                split_at = i;
                break;
            }
        }

        splitLineNoRedraw(line, split_at);
        ++line;
    }
}

/** Insert text at the current cursor position.
    \return true on success */
bool
WEditor::insertNoRedraw(string_t c)
{
    string_t line = getLine(cursor_y);
    if (line.length() < cursor_x)
        line.append(cursor_x - line.length(), ' ');

    string_t new_line = line.substr(0, cursor_x) + c;
    if (new_line.length() < max_line_length)
        new_line.append(line, cursor_x, max_line_length - new_line.length());
    setLine(cursor_y, strRTrim(new_line));

    cursor_x += c.length();
    wrapLineNoRedraw(cursor_y, max_line_length);

    /* if wrapping didn't get the cursor back in, we were at the
       rightmost column and typed a space. Move to next line. */
    if (cursor_x > max_line_length) {
        if (cursor_y < max_line-1) {
            ++cursor_y;
            insertLine(cursor_y);
            cursor_x = 0;
        } else {
            cursor_x = max_line_length;
        }
    }

    if (cursor_y >= max_line) {
        cursor_y = max_line-1;
        cursor_x = getLine(cursor_y).length();
    }
}

void
WEditor::draw(GfxCanvas& can, UpdateInfo& ui)
{
    UISkin& s = getSkin();
    s.startDraw();
    s.setSolid(true);

    for (line_t i = 0; i < max_line; ++i) {
        s.setColor(UIColor::tc_White);
        s.outTextF(can, FONT_FIXED, getExtent().x,
                   getExtent().y + i * font_heights[FONT_FIXED],
                   getExtent().w, getLine(i));
    }
    s.setColor(UIColor::tc_Yellow);
    drawVLine(can, s,
              getExtent().x + cursor_x * font_ems[FONT_FIXED],
              getExtent().y + cursor_y * font_heights[FONT_FIXED],
              getExtent().y + (cursor_y+1) * font_heights[FONT_FIXED]);

    ui.update(getExtent());
}

void
WEditor::getLayoutInfo(LayoutInfo& li)
{
    li.flags = 0;
    li.pref_size = li.min_size = GfxPoint(font_ems[FONT_FIXED] * max_line_length + 1,
                                          font_heights[FONT_FIXED] * max_line);
}

void
WEditor::doDelete(bool forward)
{
    if (!forward) {
        if (cursor_x == 0) {
            if (cursor_y <= min_line)
                return;
            --cursor_y;
            cursor_x = getLine(cursor_y).length();
        } else {
            --cursor_x;
        }
    }

    string_t line = getLine(cursor_y);
    if (cursor_x >= line.length()) {
        setLine(cursor_y, line + getLine(cursor_y + 1));
        deleteLine(cursor_y + 1);
        wrapLineNoRedraw(cursor_y, max_line_length);
    } else {
        line.erase(cursor_x, 1);
        setLine(cursor_y, line);
    }
    drawWidget();
}

bool
WEditor::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        switch (e.key.code) {
         case SDLK_UP:
            if (cursor_y > min_line) {
                select();
                --cursor_y;
                drawWidget();
            }
            return true;
         case SDLK_DOWN:
            if (cursor_y+1 < max_line) {
                select();
                ++cursor_y;
                drawWidget();
            }
            return true;
         case SDLK_LEFT:
            if (cursor_x) {
                select();
                --cursor_x;
                drawWidget();
            }
            return true;
         case SDLK_RIGHT:
            if (cursor_x < max_line_length) {
                select();
                ++cursor_x;
                drawWidget();
            }
            return true;
         case SDLK_HOME:
            select();
            cursor_x = 0;
            drawWidget();
            return true;
         case SDLK_END:
            select();
            cursor_x = getLine(cursor_y).length();
            drawWidget();
            return true;
         case SDLK_RETURN:
            select();
            if (cursor_y+1 < max_line)
                splitLineNoRedraw(cursor_y, cursor_x);
            drawWidget();
            return true;
         case SDLK_DELETE:
            select();
            doDelete(true);
            return true;
         case SDLK_BACKSPACE:
            select();
            doDelete(false);
            return true;
         case ss_Ctrl + 'y':
            select();
            deleteLine(cursor_y, 1);
            drawWidget();
            return true;
         default:
            if (e.key.unicode >= 32 && e.key.unicode != 127) {
                char c = Recoder::getInternCS().getCharacter(e.key.unicode);
                if (Recoder::isReliableChar(c)) {
                    select();
                    insertNoRedraw(string_t(1, c));
                    drawWidget();
                    return true;
                }
            }
        }
    } else if (e.type == ev_Mouse && e.mouse.b && isInside(e.mouse.x, e.mouse.y)) {
        select();
        cursor_x = (e.mouse.x - getExtent().x) / font_ems[FONT_FIXED];
        cursor_y = (e.mouse.y - getExtent().y) / font_heights[FONT_FIXED];
        drawWidget();
        return true;
    }
    return UIWidget::handleEvent(e, second_pass);
}

void
WEditor::insertString(string_t s)
{
    string_t::size_type n, p = 0;
    while ((n = s.find('\n', p)) != s.npos) {
        insertNoRedraw(s.substr(p, n - p));
        if (cursor_y+1 < max_line)
            splitLineNoRedraw(cursor_y, cursor_x);
        p = n+1;
    }
    insertNoRedraw(s.substr(p));
}
