/**
  *  \file game/sim-game.cc
  *  \brief Simulator / Game Interface
  *
  *  This is the "real" interface to a game, which links the simulator
  *  UI to real game data. This allows people to add ships from game
  *  to simulator and vice versa.
  */

#include "game/sim-game.h"

#include <algorithm>
#include "game/planet.h"
#include "game/ship.h"
#include "game/mission.h"
#include "game/cargo.h"

GSimulatorRealGameInterface::GSimulatorRealGameInterface(GUniverse& univ)
    : univ(univ)
{ }

GSimulatorRealGameInterface::~GSimulatorRealGameInterface()
{ }

string_t
GSimulatorRealGameInterface::getPlanetName(int id)
{
    return univ.getPlanet(id).getName();
}

bool
GSimulatorRealGameInterface::isGameAvailable()
{
    return true;
}

/** Check ship's relation to the game.
    \todo Handle targets. */
GSimulatorRealGameInterface::ShipRelation
GSimulatorRealGameInterface::getShipRelation(const GSimShip& sh)
{
    if (!univ.getShipType().isExistant(sh.getId()))
        return ShipIsUnknown;

    GShip& game_ship = univ.getShip(sh.getId());
    if (sh.getOwner() != game_ship.getRealOwner())
        return ShipIsUnknown;
    if (sh.getHull() != game_ship.getHullId())
        return ShipIsUnknown;

    return ShipIsPlayed;
}

bool
GSimulatorRealGameInterface::selectGameShip(const GSimShip& sh)
{
    if (!univ.getShipType().isExistant(sh.getId()))
        return false;

    univ.getShipType().setCurrent(sh.getId());
    return true;
}

bool
GSimulatorRealGameInterface::updateFromGame(GSimShip& sh)
{
    if (!univ.getShipType().isExistant(sh.getId()))
        return false;

    const GShip& game_ship = univ.getShip(sh.getId());

    if (game_ship.haveData()) {
        sh.setCrew(game_ship.getCrew());
        sh.setHull(game_ship.getHullId());
        sh.setBeamType(game_ship.getBeamType());
        sh.setBeamCount(game_ship.getBeamCount());
        sh.setTorpType(game_ship.getTorpType());
        sh.setTorpLauncherCount(game_ship.getTorpLauncherCount());
        sh.setBayCount(game_ship.getBayCount());
        sh.setAmmo(game_ship.getAmmo());
        sh.setEngineType(game_ship.getEngine());

        if (game_ship.getCargo(el_Neutronium) == 0)
            sh.setAggressiveness(sh.agg_NoFuel);
        else if (game_ship.getMission() == GMission::msn_Kill)
            sh.setAggressiveness(sh.agg_Kill);
        else
            sh.setAggressiveness(game_ship.getPE());

        sh.setId(game_ship.getId());
        sh.setName(game_ship.getName());
        sh.setFCode(game_ship.getFCode());
        sh.setDamage(std::max(int(game_ship.getDamage() - game_ship.getCargo(el_Supplies) / 5), 0));
        sh.setShield(std::max(int(100 - sh.getDamage()), 0));
        sh.setOwner(game_ship.getRealOwner());
        sh.setExperienceLevel(0); // FIXME

        sh.setFlags(0);
        if (isMissionCloaking(game_ship.getMission(), game_ship.getRealOwner())
            && game_ship.canDoSpecial(hf_Cloak)
            && game_ship.getCargo(el_Neutronium) > 0)
        {
            sh.setFlags(sh.getFlags() | GSimShip::fl_Cloaked);
        }
        return true;
    } else {
        // FIXME: handle partial data
        return false;
    }
}

bool
GSimulatorRealGameInterface::updateToGame(const GSimShip& sh)
{
    if (!univ.getShipType().isExistant(sh.getId()))
        return false;

    GShip& game_ship = univ.getShip(sh.getId());
    // can't modify crew, hull, beam type, beam count, torp type,
    // launcher count, bay count, engine type, Id

    // Mission
    // FIXME: check whether we're allowed to modify the mission
    if (sh.getAggressiveness() == GSimShip::agg_Kill) {
        game_ship.setMission(GMission::msn_Kill, 0, 0);
    } else {
        if (game_ship.canDoSpecial(hf_Cloak) && (sh.getFlags() & GSimShip::fl_Cloaked) != 0) {
            if (!isMissionCloaking(game_ship.getMission(), game_ship.getRealOwner()))
            {
                game_ship.setMission(GMission::msn_Cloak, 0, 0);
            }
        } else {
            if (isMissionCloaking(game_ship.getMission(), game_ship.getRealOwner())
                || game_ship.getMission() == GMission::msn_Kill)
            {
                game_ship.setMission(0, 0, 0);
            }
        }
    }

    // PE
    if (sh.getAggressiveness() > 0 && sh.getAggressiveness() <= NUM_PLAYERS)
        game_ship.setPE(sh.getAggressiveness());
    else
        game_ship.setPE(0);

    // FCode, Name
    game_ship.setFCode(sh.getFCode());
    game_ship.setName(sh.getName());

    // Ammo
    // FIXME: in case we handle mkt/lfm someday, we would have to revert it here
    int pid;
    if (sh.getAmmo() != game_ship.getAmmo()
        && (pid = game_ship.getOrbit()) != 0
        && univ.getPlanetType().isExistant(pid))
    {
        GPlanet& game_planet = univ.getPlanet(pid);
        if (game_planet.getOwner() == game_ship.getOwner()) {
            GCargoType type;
            if (sh.getTorpLauncherCount() != 0)
                type = getCargoTypeFromTorpType(sh.getTorpType());
            else if (sh.getBayCount() != 0)
                type = el_Fighters;
            else
                type = el_Neutronium;

            if (type != el_Neutronium) {
                // use regular cargo transfer
                GMultiTransaction mt;
                GCargoContainer& pt = mt.addNewSubtransaction(new GPlanetTransfer(game_planet));
                GCargoContainer& st = mt.addNewSubtransaction(new GShipTransfer(game_ship));
                GCargoTransfer& ct = mt.addNewSubtransaction(new GCargoTransfer(pt, st));
                ct.move(type, sh.getAmmo() - game_ship.getAmmo(), false, true, false);
                mt.doCommit();
            }
        }
    }

    return true;
}

bool
GSimulatorRealGameInterface::isGamePlanet(const GSimPlanet& pl)
{
    if (!univ.getPlanetType().isExistant(pl.getId()))
        return false;

    GPlanet& game_planet = univ.getPlanet(pl.getId());
    if (pl.getOwner() != game_planet.getOwner())
        return false;

    return true;
}

bool
GSimulatorRealGameInterface::updateFromGame(GSimPlanet& pl)
{
    // FIXME: this function isn't yet used by the UI, hence not yet implemented
    (void) pl;
    return false;
}

bool
GSimulatorRealGameInterface::updateToGame(const GSimPlanet& pl)
{
    // FIXME: this function isn't yet used by the UI, hence not yet implemented
    (void) pl;
    return false;
}
