/**
  *  \file int/exec.h
  *
  *  (c) 2001-2006 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_INT_EXEC_H
#define PCC_V2_INT_EXEC_H

#include <vector>
#include "int/data.h"
#include "int/bytecode.h"
#include "int/vars.h"
#include "cpluslib/smartptr.h"

class IntLexer;

struct IntFrame {
    /** The BCO running in this frame. */
    IntByteCodeObject&          bco;
    /** Instruction pointer (next insn to execute). */
    IntByteCodeObject::index_t  ip;

    /** Names of local variables. */
    IntVariableNames            local_names;
    /** Values of local variables. */
    IntVariableValues           local_values;
    /** Context-TOS for this frame. All contexts at index >= context_tos
        belong to this frame. */
    std::size_t                 context_tos;

    IntFrame(IntByteCodeObject& bco, std::size_t context_tos);
    IntFrame(const IntFrame& rhs);
    ~IntFrame();
    IntFrame* clone() const;
 private:
    IntFrame& operator=(const IntFrame&);
};

class IntExecutionContext {
    IntExecutionContext& operator=(const IntExecutionContext&);

    struct Exception {
        std::size_t context_tos, control_tos, value_tos;
        IntByteCodeObject::index_t ip;
        Exception(IntExecutionContext&, IntByteCodeObject::index_t);
    };
    friend class Exception;

    typedef std::vector<Ptr<IntContext> > context_t;
    typedef std::vector<Ptr<IntFrame> >   control_stack_t;
    typedef std::vector<IntObject>        value_stack_t;
    typedef std::vector<Exception>        exception_stack_t;

    context_t         context;
    control_stack_t   control_stack;
    value_stack_t     value_stack;
    exception_stack_t exception_stack;

    int parseArgs(IntLexer& lex, IntByteCodeObject& bco);

 public:
    IntExecutionContext();
    IntExecutionContext(const IntExecutionContext& rhs);
    ~IntExecutionContext();

    void pushContext(Ptr<IntContext> con);
    void popContext();

    void pushFrame(Ptr<IntFrame> f);
    // void popFrame();

    void doException(string_t s);
    void doReturn();
    void runOnce();
    void run();

    const context_t&        getContext() const { return context; }
    const control_stack_t&  getControlStack() const { return control_stack; }
    const value_stack_t&    getValueStack() const { return value_stack; }

    IntObject peekV(int index) const { return value_stack[value_stack.size()-index-1]; }
    IntObject popV()                 { IntObject tmp = value_stack.back(); value_stack.pop_back(); return tmp; }
    void      pushV(IntObject o)     { value_stack.push_back(o); }
    void      popVs(int count)       { value_stack.erase(value_stack.end() - count, value_stack.end()); }

    enum Status {
        ps_Running,             ///< Process is running.
        ps_Runnable,            ///< Process is ready.
        ps_Stopped,             ///< Stopped.
        ps_Disabled,            ///< Stopped, but user doesn't want it to run.
        ps_Terminated,          ///< Terminated, but could be continued.
        ps_Finished,            ///< Terminated, can't be continued.
        ps_Frozen,              ///< Frozen, i.e.\ being examined from the outside.

        ps_LAST = ps_Frozen
    };

    Status    getStatus() const { return status; }
    void      setStatus(Status st) { status = st; }

    const string_t& getName() const { return name; }
    void            setName(const string_t& n) { name = n; }

    int       getPriority() const { return priority; }
    void      setPriority(int n);
 private:
    Status    status;
    string_t  name;
    int       priority;

    void      doEval(string_t s);
};

class IntProcessList {
    // FIXME: I'm not happy with this class using smart pointers
    typedef std::vector<Ptr<IntExecutionContext> > process_list_t;

    process_list_t list;
 public:
    IntProcessList();
    ~IntProcessList();
    void runRunnableProcesses();
    void runAllProcesses();
    void addProcess(Ptr<IntExecutionContext> process);
    void killTerminatedProcesses();

    void reschedule(Ptr<IntExecutionContext> process, int pri);
};

// Functions which are called by the interface compiler
int32        IF_intValue(const IntObject& o, const char* name);
bool         IF_boolValue(const IntObject& o, const char* name);
double       IF_floatValue(const IntObject& o, const char* name);
string_t     IF_stringValue(IntObject o, const char* name);
bool         IF_argCheck(int got_args, int min_args, int max_args,
                         const char* name, IntExecutionContext& xc);

#endif
